/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */
package jorm.rdb.mysql;

// Import clauses required by jorm
import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.jorm.interp.genclass.api.FieldDesc;
import org.objectweb.jorm.interp.genclass.api.GenClassProp;
//import for the rdb mapper
import org.objectweb.transaction.api.ConnectionWrapper;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.PreparedStatement;
import java.sql.Statement;
import java.util.HashMap;
import java.util.NoSuchElementException;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.interp.genclass.rdb.api.RdbFieldDesc;
import org.objectweb.jorm.interp.genclass.rdb.api.RdbGenClassProp;

import java.util.Date;
import java.util.Iterator;

public class Cmp2_CustomerMapping
	extends org.objectweb.jonas_ejb.container.jorm.RdbFactory
	implements PClassMapping, Loggable {

	protected static final String MAPPER_NAME = "rdb.mysql";
	protected static final String MAPPING_NAME = "jonas";
	//Fields
	protected static final String lastName_NAME = "lastName";
	protected static final String firstName_NAME = "firstName";
	protected static final String hasGoodCredit_NAME = "hasGoodCredit";
	protected static final String id_NAME = "id";
	protected static final String creditCard_NAME = "creditCard";
	protected static final String homeAddress_NAME = "homeAddress";
	protected static final String reservations_NAME = "reservations";
	protected static final String phoneNumbers_NAME = "phoneNumbers";
	//Hidden fields
	protected static final String creditCard_id_NAME = "creditCard_id";
	protected static final String homeAddress_id_NAME = "homeAddress_id";


	private final static String[][] SUPER_TYPES =  {
    };


	/** Fully quailied name of the persitent class */
	protected final static String JORM_CLASS_NAME =
		"jorm.Cmp2_Customer";

	/** The actual binder which does the binding action */
	private PBinder binder = null;

	/** The meta object of the current jorm class */
	private Class moc = null;

	/** The mapper linked to this PClassMapping */
	private PMapper mapper = null;

	PNamingContext creditCardPnc = null;
	boolean creditCardCodingOnly = true;

	PNamingContext homeAddressPnc = null;
	boolean homeAddressCodingOnly = true;

	PNamingContext reservationsPnc = null;
	boolean reservationsCodingOnly = true;
	PClassMapping reservationsGCM = null;

	PNamingContext phoneNumbersPnc = null;
	boolean phoneNumbersCodingOnly = true;
	PClassMapping phoneNumbersGCM = null;

 
	/**
	 * The logger instance used by this XMapping instance
	 * and the XBinding instances
	 */
	protected Logger logger = null;
	protected LoggerFactory loggerFactory = null;

	public Cmp2_CustomerMapping() {
		super();
 	}


	// IMPLEMENTATION OF METHODS FROM THE PClassMapping INTERFACE

	public void init(PMapper m, MetaObject metaclass) throws PException {
		if (Debug.ON && logger!=null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "");
		mapper = m;
		moc = (Class) metaclass;
		PType ptype = mapper.getPTypeSpace().createPType(
			"jorm.Cmp2_Customer",
			SUPER_TYPES);
		setPType(ptype);
		String[][] creditCardICDef = {
		};
		ptype = mapper.getPTypeSpace().createPType(
			"jorm.Cmp2_CreditCard",
			creditCardICDef);
		creditCardPnc.setPType(ptype);
		String[][] homeAddressICDef = {
		};
		ptype = mapper.getPTypeSpace().createPType(
			"jorm.Cmp2_Address",
			homeAddressICDef);
		homeAddressPnc.setPType(ptype);
		FieldDesc[] fds = null;
  		fds = new FieldDesc[1];
		fds[0] = new RdbFieldDesc("Cmp2_Customer_id", "null", PTypeSpace.OBJINT, "Cmp2_Customer_id");
		((GenClassProp) reservationsGCM ).setIdentifierDesc(fds);
		fds = new FieldDesc[0];
 		((GenClassProp) reservationsGCM ).setIndexDesc(fds);
		fds = new FieldDesc[2];
		fds[0] = new RdbFieldDesc(null, null, PTypeSpace.REFTOP, null);
		fds[1] = new RdbFieldDesc("Cmp2_Reservation_id", "null", PTypeSpace.OBJINT, "Cmp2_Reservation_id");
 		((GenClassProp) reservationsGCM).setElemDesc(fds);
		((RdbGenClassProp) reservationsGCM).setTableName("CMP2_RESERVATION_CMP2_CUSTOMER");
		((RdbGenClassProp) reservationsGCM).setTypeConverter(
			new org.objectweb.jorm.runtime.mapper.rdb.adapter.lib.MysqlAdapter());
 		String[] reservationsGCDef = {"java.util.Collection"};
		String[][] reservationsICDef = {
		};
		ptype = mapper.getPTypeSpace().createPType(
			reservationsGCDef,
			"jorm.Cmp2_Reservation",
			reservationsICDef);
		((GenClassProp) reservationsGCM).setGcType(ptype);
  		fds = new FieldDesc[1];
		fds[0] = new RdbFieldDesc("Cmp2_Customer_id", "null", PTypeSpace.OBJINT, "Cmp2_Customer_id");
		((GenClassProp) phoneNumbersGCM ).setIdentifierDesc(fds);
		fds = new FieldDesc[0];
 		((GenClassProp) phoneNumbersGCM ).setIndexDesc(fds);
		fds = new FieldDesc[2];
		fds[0] = new RdbFieldDesc(null, null, PTypeSpace.REFTOP, null);
		fds[1] = new RdbFieldDesc("Cmp2_Phone_id", "null", PTypeSpace.OBJINT, "Cmp2_Phone_id");
 		((GenClassProp) phoneNumbersGCM).setElemDesc(fds);
		((RdbGenClassProp) phoneNumbersGCM).setTableName("CMP2_PHONE_CMP2_CUSTOMER");
		((RdbGenClassProp) phoneNumbersGCM).setTypeConverter(
			new org.objectweb.jorm.runtime.mapper.rdb.adapter.lib.MysqlAdapter());
 		String[] phoneNumbersGCDef = {"java.util.Collection"};
		String[][] phoneNumbersICDef = {
		};
		ptype = mapper.getPTypeSpace().createPType(
			phoneNumbersGCDef,
			"jorm.Cmp2_Phone",
			phoneNumbersICDef);
		((GenClassProp) phoneNumbersGCM).setGcType(ptype);
	}

	public PMapper getPMapper() {
		return mapper;
	}

	public PBinder getPBinder() {
		return binder;
	}

	/** This method is used only in the case of the current
	 * class is a gen class and contains references*/
	public PNamingContext getPNameManager()
		throws UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
	}

	public PNamingContext getPNameManager(String fn)
		throws UnsupportedOperationException {
		if (creditCard_NAME.equals(fn))
			return creditCardPnc;
		if (homeAddress_NAME.equals(fn))
			return homeAddressPnc;
		if (reservations_NAME.equals(fn))
			return reservationsPnc;
		if (phoneNumbers_NAME.equals(fn))
			return phoneNumbersPnc;
		throw new UnsupportedOperationException("This class does'nt contain field reference");
	}

	public String getClassName() {
		return JORM_CLASS_NAME;
	}

	/** This method is used only in the case of the current
	 * class is a gen class and contains another gen class*/
	public PClassMapping getGenClassMapping() throws UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
	}

	public PClassMapping getGenClassMapping(String fn)
		throws UnsupportedOperationException {
		if (reservations_NAME.equals(fn))
			return reservationsGCM;
		if (phoneNumbers_NAME.equals(fn))
			return phoneNumbersGCM;
		throw new UnsupportedOperationException("This class does not contain a " + fn + " field which would reference a generic class");
	}

	public boolean isConform(String mapperName) {
		return MAPPER_NAME.equals(mapperName);
	}

	public String getMappingName() {
		return MAPPING_NAME;
	}

	public MetaObject getMetaInfo() {
		return moc;
	}

	public String[][] getSuperClassNames() {
		return SUPER_TYPES;
	}

	public void setPBinder(PBinder b) {
		binder = b;
	}

	/** This method is used only in the case of the current
	 * class is a gen class and contains references*/
	public void setPNamingContext(PNamingContext nc, boolean co) throws PException, UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
	}

	public void setGenClassMapping(PClassMapping pcm) throws UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
	}

	/** This method method must check if the
	  * the fn parameter represents a reference field
	  * Then register the association field name / nc
	  * into the hashMap.
	  * If the jorm class does'nt contains field reference
	  * then an UnsupportedOperationException is throwed */
    public void setPNamingContext(String fn, PNamingContext nc, boolean co)
		throws PException, UnsupportedOperationException {

		// Checking if fn designes a reference field
		if (creditCard_NAME.equals(fn)) {
			// checking if the PNamingContext is able to encode
 			if (!nc.codingSupported(PNamingContext.CTOINT)) {
				throw new PExceptionTyping("The " + fn
					+ " field doesn't support the Int encoding");
			}
			creditCardPnc = nc;
			creditCardCodingOnly = co;
			return;
		}
		if (homeAddress_NAME.equals(fn)) {
			// checking if the PNamingContext is able to encode
 			if (!nc.codingSupported(PNamingContext.CTOINT)) {
				throw new PExceptionTyping("The " + fn
					+ " field doesn't support the Int encoding");
			}
			homeAddressPnc = nc;
			homeAddressCodingOnly = co;
			return;
		}
		if (reservations_NAME.equals(fn)) {
			// checking if the PNamingContext is able to encode
 			if (!nc.codingSupported(PNamingContext.CTOINT)) {
				throw new PExceptionTyping("The " + fn
					+ " field doesn't support the Int encoding");
			}
			reservationsPnc = nc;
			reservationsCodingOnly = co;
			return;
		}
		if (phoneNumbers_NAME.equals(fn)) {
			// checking if the PNamingContext is able to encode
 			if (!nc.codingSupported(PNamingContext.CTOINT)) {
				throw new PExceptionTyping("The " + fn
					+ " field doesn't support the Int encoding");
			}
			phoneNumbersPnc = nc;
			phoneNumbersCodingOnly = co;
			return;
		}
		throw new UnsupportedOperationException(
			"No field reference with the name " + fn + " in this class");
	}

	public void setGenClassMapping(String fn, PClassMapping pcm) throws UnsupportedOperationException {
		if (reservations_NAME.equals(fn)) {
			reservationsGCM = pcm;
			return;
		}
		if (phoneNumbers_NAME.equals(fn)) {
			phoneNumbersGCM = pcm;
			return;
		}
		throw new UnsupportedOperationException("No field reference to a generic class with the name " + fn + " in this class");
	}

	public PBinding createPBinding() throws PException {
		if (Debug.ON && logger!=null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "");
		PBinding binding = new Cmp2_CustomerBinding();
		try {
			binding.init(this);
		}
		catch(PExceptionProtocol pep) {
		}
		return binding;
	}
	public void initMappingStructures(Object conn, byte cleanup)
		throws PException {
		if (Debug.ON && logger!=null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "");

		Connection connection = null;
		if (conn==null)
			throw new PExceptionProtocol("Impossible to initialize the mapping structures without connection");
		else if (conn instanceof Connection)
			connection = (Connection) conn;
		else if (conn instanceof ConnectionWrapper)
			connection = (Connection) ((ConnectionWrapper) conn).getNonCCIConnection();
		else
			throw new PExceptionProtocol("Impossible to initialize the mapping structures with this connection type.");
		try {
			//Ask to the database if the table already exists
			// Be carefull some driver does not support the table name partten,
			// then the table name is searched in the resultSet.
			ResultSet rs = connection.getMetaData().getTables(null,null, "CMP2_CUSTOMER_" ,null);
			boolean existTable = false;
			while(rs.next() && !existTable)
				existTable = "CMP2_CUSTOMER_".equalsIgnoreCase(rs.getString(3));
			rs.close();
			if (existTable) {
				if (cleanup==CLEANUP_REMOVEDATA)
					removeMappingStructures(conn, true);
				else if (cleanup==CLEANUP_REMOVEALL) {
					removeMappingStructures(conn, false);
					existTable = false;
				}
			}
			if (!existTable) {
				Statement stmt = connection.createStatement();
				stmt.execute("CREATE TABLE CMP2_CUSTOMER_ (id_ INTEGER NOT NULL, lastName_ VARCHAR(254), firstName_ VARCHAR(254), hasGoodCredit_ BIT, creditCard_id INTEGER, homeAddress_id INTEGER, PRIMARY KEY (id_))");
				stmt.close();
			}
		}
		catch(java.sql.SQLException se) {
			throw new PExceptionIO(se);
		}
		reservationsGCM.initMappingStructures(conn, cleanup);
		phoneNumbersGCM.initMappingStructures(conn, cleanup);
	}
	public void removeMappingStructures(Object conn, boolean contentonly)
		throws PException {
		if (Debug.ON && logger!=null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "");
		Connection connection = null;
		if (conn==null)
			throw new PExceptionProtocol("Impossible to initialize the mapping structures without connection");
		else if (conn instanceof Connection)
			connection = (Connection) conn;
		else if (conn instanceof ConnectionWrapper)
			connection = (Connection) ((ConnectionWrapper) conn).getNonCCIConnection();
		else
			throw new PExceptionProtocol("Impossible to initialize the mapping structures with this connection type.");
		try {
			Statement stmt = connection.createStatement();
			if (contentonly)
				// Remove all data only
				stmt.execute("DELETE FROM CMP2_CUSTOMER_");
			else
				// Remove all data and the structures
				stmt.execute("DROP TABLE CMP2_CUSTOMER_");
			stmt.close();
		}
		catch(java.sql.SQLException se) {
			throw new PExceptionIO(se);
		}
		reservationsGCM.removeMappingStructures(conn, contentonly);
		phoneNumbersGCM.removeMappingStructures(conn, contentonly);
	}

    public Iterator getPNameIterator(Object conn) throws PException {
                Connection connection = null;
		if (conn==null)
			throw new PExceptionProtocol("Impossible to initialize the mapping structures without connection");
		else if (conn instanceof Connection)
			connection = (Connection) conn;
		else if (conn instanceof ConnectionWrapper)
			connection = (Connection) ((ConnectionWrapper) conn).getNonCCIConnection();
		else
			throw new PExceptionProtocol("Impossible to initialize the mapping structures with this connection type.");
		try {
			PreparedStatement ps = connection.prepareStatement(
                "SELECT id_ FROM CMP2_CUSTOMER_");
            return new Cmp2_CustomerResultIterator(ps, ps.executeQuery(), binder);
        } catch (java.sql.SQLException e) {
            throw new PExceptionIO(e);
        }
    }

    public void configureRefFields(ReferenceConfigurator rc) throws PException, UnsupportedOperationException {
        PClassMapping gcm = null;
        boolean isco;
		setPNamingContext(
            creditCard_NAME,
            rc.getPNameManager(JORM_CLASS_NAME, creditCard_NAME, "jorm.Cmp2_CreditCard"),
            rc.isCodingOnly(JORM_CLASS_NAME, creditCard_NAME));
		setPNamingContext(
            homeAddress_NAME,
            rc.getPNameManager(JORM_CLASS_NAME, homeAddress_NAME, "jorm.Cmp2_Address"),
            rc.isCodingOnly(JORM_CLASS_NAME, homeAddress_NAME));
        isco = rc.isCodingOnly(JORM_CLASS_NAME, reservations_NAME);
        setPNamingContext(
                reservations_NAME,
                rc.getPNameManager(JORM_CLASS_NAME, reservations_NAME, new String[]{"java.util.Collection"}),
                isco);
        setGenClassMapping(
                reservations_NAME,
                rc.getGenClassMapping(JORM_CLASS_NAME, reservations_NAME, new String[]{"java.util.Collection"}));
        isco = rc.isCodingOnly(JORM_CLASS_NAME, phoneNumbers_NAME);
        setPNamingContext(
                phoneNumbers_NAME,
                rc.getPNameManager(JORM_CLASS_NAME, phoneNumbers_NAME, new String[]{"java.util.Collection"}),
                isco);
        setGenClassMapping(
                phoneNumbers_NAME,
                rc.getGenClassMapping(JORM_CLASS_NAME, phoneNumbers_NAME, new String[]{"java.util.Collection"}));
    }

	public void setPType(PType pt) {
		binder.setPType(pt);
	}


	// IMPLEMENTATION OF METHODS FROM THE Loggable INTERFACE

	public void setLogger(Logger l) {
		logger = l;
		((Loggable) reservationsGCM).setLogger(l);
		((Loggable) phoneNumbersGCM).setLogger(l);
	}

	public void setLoggerFactory(LoggerFactory lf) {
		loggerFactory = lf;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}


}
class Cmp2_CustomerResultIterator implements Iterator
        {
        ResultSet rs;
		PBinder pnc;
        boolean hasNext;
        boolean isForw;
		PreparedStatement ps;
        Cmp2_CustomerResultIterator(PreparedStatement ps, ResultSet r,  PBinder pnc)
			throws java.sql.SQLException {
			this.ps = ps;
			this.pnc = pnc;
            rs = r;
            hasNext = rs.next();
            isForw = true;
        }

		public void finalize() throws Throwable {
			if (rs != null)
				rs.close();
			if (ps != null)
				ps.close();
			super.finalize();
		}

        public boolean hasNext() {
            if (isForw) {
                return hasNext;
            }
            else {
                isForw = true;
                try {
                    return hasNext = rs.next();
                } catch(Exception e) {
                    e.printStackTrace();
                    return false;
                }
            }
        }

        public Object next() throws NoSuchElementException {
            try {
                if (!isForw) {
                    hasNext = rs.next();
                    isForw = true;
                }
                if (!hasNext) {
                    throw new NoSuchElementException();
                }
                isForw = false;
                return pnc.decodeOint( new Integer(rs.getInt(1)));
            } catch (Exception e) {
                throw new NoSuchElementException(
                    "Impossible to decode to a PName: " + e.getMessage());
            }
        }

        public void remove()
            throws UnsupportedOperationException, IllegalStateException {
            throw new UnsupportedOperationException("");
        }

}
